/* Non pre-empting synchronization kernel, using LWP */
#include <stdio.h>
#include <lwp/lwp.h>
#include <lwp/lwpmachdep.h>
#include <lwp/lwperror.h>
#include "global.h"
#include "mbuf.h"
#include "proc.h"
#include "timer.h"
#include "socket.h"
#include "daemon.h"

struct proc *Curproc;		/* Currently running process */
struct proc *Rdytab;		/* Processes ready to run (not including curproc) */
struct proc *Waittab[PHASH];	/* Waiting process list */
static struct mbuf *Killq;

static void addproc __ARGS((struct proc *entry));
static void delproc __ARGS((struct proc *entry));
static void startproc __ARGS((int pp, int pc, int iarg, int parg1, int parg2));

/* Create a process descriptor for the main function. Must be actually
 * called from the main function!
 */
struct proc *
mainproc(name)
char *name;
{
	register struct proc *pp;

	/* Create process descriptor */
	pp = (struct proc *)callocw(1,sizeof(struct proc));

	/* Find our own thread ID */
	lwp_self(&pp->tid);
	
	/* Create name */
	pp->name = strdup(name);
	pp->stksize = 0;
	pp->output = Console;

        /* Task initially runs with interrupts on */
        pp->i_state = 1;

	/* Make current */
	pp->state = READY;
	Curproc = pp;

	pod_setmaxpri(MINPRIO+1);	/* allow two priority levels */

	return pp;
}
/* Create a new, ready process and return pointer to descriptor.
 * The general registers are not initialized, but optional args are pushed
 * on the stack so they can be seen by a C function.
 */
struct proc *
newproc(name,stksize,pc,iarg,parg1,parg2)
char *name;		/* Arbitrary user-assigned name string */
unsigned int stksize;	/* Stack size in words to allocate */
void (*pc)();		/* Initial execution address */
int iarg;		/* Integer argument (argc) */
void *parg1;		/* Generic pointer argument #1 (argv) */
void *parg2;		/* Generic pointer argument #2 (session ptr) */
{
	register struct proc *pp;
	int i;
	void startproc();
	
	/* Create process descriptor */
	pp = (struct proc *)callocw(1,sizeof(struct proc));

	/* Create name */
	pp->name = strdup(name);

	/* Allocate stack */
#if defined(AMIGA) || defined(UNIX)
	stksize += 2000;	/* DOS overhead */
#endif
	if(stksize < MINSTACKSZ)
	     stksize = MINSTACKSZ;
     	pp->stksize = stksize;
/*
	if((pp->stack = (stkalign_t *)malloc(sizeof(stkalign_t)*stksize)) == NULL){
		free(pp->name);
		free((char *)pp);
		return NULLPROC;
	}
*/
	pp->stack = lwp_newstk(); /* Grab a preallocated stack */
	if(lwp_create(&pp->tid, startproc, MINPRIO, 0, pp->stack, 5,
		      (int)pp,(int)pc,iarg,(int)parg1,(int)parg2) == -1) {
	     lwp_perror("lwp_create");
	     free(pp->name);
	     free((char *)pp);
	     return NULLPROC;
	}
	
	/* Initialize stack for high-water check */
	lwp_stkcswset(pp->tid, (caddr_t) pp->stack - 8192);

        /* Task initially runs with interrupts on */
        pp->i_state = 1;

	/* Inherit creator's input and output sockets */
	usesock(Curproc->input);
	pp->input = Curproc->input;
	usesock(Curproc->output);
	pp->output = Curproc->output;

        /* Add to ready process table */
        pp->state = READY;
        addproc(pp);

	return pp;
}

static void
startproc(pp, pc, iarg, parg1, parg2)
int pp, pc, iarg, parg1, parg2;
{
     void (*func)();
     func = (void (*)()) pc;
     Curproc = (struct proc *) pp;
     delproc (Curproc);
     (*func)(iarg,(void *)parg1,(void *)parg2);
     killself();
}

/* Free resources allocated to specified process. If a process wants to kill
 * itself, the reaper is called to do the dirty work. This avoids some
 * messy situations that would otherwise occur, like freeing your own stack.
 */
void
killproc(pp)
register struct proc *pp;
{
	if(pp == NULLPROC)
		return;
	/* Don't check the stack here! Will cause infinite recursion if
	 * called from a stack error
	 */

	if(pp == Curproc)
		killself();	/* Doesn't return */

	/* Close any open sockets */
	freesock(pp);

	close_s(pp->input);
	close_s(pp->output);

	/* Stop alarm clock in case it's running */
	stop_timer(&pp->alarm);

	/* Alert everyone waiting for this proc to die */
	psignal(pp,0);

	/* Remove from appropriate table */
	delproc(pp);

	/* Free allocated memory resources */
	free(pp->name);
	free(pp->outbuf);

	lwp_destroy(pp->tid);
	free((char *)pp);
}

/* Terminate current process by sending a request to the killer process.
 * Automatically called when a process function returns. Does not return.
 */
void
killself()
{
	register struct mbuf *bp;
	bp = pushdown(NULLBUF,sizeof(Curproc));
	memcpy(bp->data,(char *)&Curproc,sizeof(Curproc));
	enqueue(&Killq,bp);
	/* "Wait for me; I will be merciful and quick." */
	for(;;)
		lwp_suspend(SELF);
}

/* Process used by processes that want to kill themselves */
void
killer(i,v1,v2)
int i;
void *v1;
void *v2;
{
	struct proc *pp;
	struct mbuf *bp;

	for(;;){
		while(Killq == NULLBUF)
			pwait(&Killq);
		bp = dequeue(&Killq);
		pullup(&bp,(char *)&pp,sizeof(pp));
		free_p(bp);
		if(pp != Curproc)	/* We're immortal */
			killproc(pp);
	}						
}

/* Wakeup waiting process, regardless of event it's waiting for. The process
 * will see a return value of "val" from its pwait() call.
 */
void
alert(pp,val)
struct proc *pp;
int val;
{
	if(pp == NULLPROC)
		return;
#ifdef	notdef
	if((pp->state & WAITING) == 0)
		return;
#endif
	if(pp != Curproc)
		delproc(pp);
	pp->state &= ~WAITING;
	pp->retval = val;
	pp->event = 0;
	if(pp != Curproc)
		addproc(pp);
	lwp_resume(pp->tid);
}

/* Post a wait on a specified event and give up the CPU until it happens. The
 * null event is special: it means "I don't want to block on an event, but let
 * somebody else run for a while". It can also mean that the present process
 * is terminating; in this case the wait never returns.
 *
 * Pwait() returns 0 if the event was signaled; otherwise it returns the
 * arg in an alert() call. Pwait must not be called from interrupt level.
 *
 */
int
pwait(event)
void *event;
{
	register struct proc *oldproc;
	int tmp;
	struct proc *pp;
	
	pp = Curproc;
	pp->i_state = istate();
	if(event != NULL) {
	     /* Post a wait for the specified event */
	     Curproc->event = event;
	     Curproc->state = WAITING;
	}
	addproc(Curproc);
	restore(1);
	if(Curproc->event == NULL)
	     /* Special case; just give up the processor. */
	     lwp_yield(SELF);
	else
	     lwp_suspend(SELF);
	Curproc = pp;
	delproc(Curproc);
	/* At this point, we're running in the newly dispatched task */
	tmp = Curproc->retval;
	Curproc->retval = 0;
	dirps();
	restore(Curproc->i_state);
	return tmp;
}

/* Make ready the first 'n' processes waiting for a given event. The ready
 * processes will see a return value of 0 from pwait().  Note that they don't
 * actually get control until we explicitly give up the CPU ourselves through
 * a pwait(). Psignal may be called from interrupt level. It returns the
 * number of processes that were woken up.
 */
int
psignal(event,n)
void *event;	/* Event to signal */
int n;		/* Max number of processes to wake up */
{
	register struct proc *pp;
	struct proc *pnext;
	int i_state;
	unsigned int hashval;
	int cnt = 0;

	if(event == NULL)
		return 0;		/* Null events are invalid */

	/* n = 0 means "signal everybody waiting for this event" */
	if(n == 0)
		n = 65535;

	hashval = phash(event);
	i_state = dirps();
	for(pp = Waittab[hashval];n != 0 && pp != NULLPROC;pp = pnext){
		pnext = pp->next;
		if(pp->event == event){
			delproc(pp);
			pp->state &= ~WAITING;
			pp->retval = 0;
			pp->event = NULL;
			addproc(pp);
			lwp_resume(pp->tid);
			n--;
			cnt++;
		}
	}
	restore(i_state);
	return cnt;
}

/* Rename a process */
void
chname(pp,newname)
struct proc *pp;
char *newname;
{
	free(pp->name);
	pp->name = strdup(newname);
}
/* Remove a process entry from the appropriate table */
static void
delproc(entry)
register struct proc *entry;	/* Pointer to entry */
{
	int i_state;

	if(entry == NULLPROC)
		return;

	i_state = dirps();
	if(entry->next != NULLPROC)
		entry->next->prev = entry->prev;
	if(entry->prev != NULLPROC){
		entry->prev->next = entry->next;
	} else {
		switch(entry->state){
		case READY:
			Rdytab = entry->next;
			break;
		case WAITING:
			Waittab[phash(entry->event)] = entry->next;
			break;
		}
	}
	restore(i_state);
}
/* Append proc entry to end of appropriate list */
static void
addproc(entry)
register struct proc *entry;	/* Pointer to entry */
{
	register struct proc *pp;
	struct proc **head;
	int i_state;

	if(entry == NULLPROC)
		return;

	i_state = dirps();
	switch(entry->state){
	case READY:
		head = &Rdytab;
		break;
	case WAITING:
		head = &Waittab[phash(entry->event)];
		break;
	}
	entry->next = NULLPROC;
	if(*head == NULLPROC){
		/* Empty list, stick at beginning */
		entry->prev = NULLPROC;
		*head = entry;
	} else {
		/* Find last entry on list */
		for(pp = *head;pp->next != NULLPROC;pp = pp->next)
			;
		pp->next = entry;
		entry->prev = pp;
	}
	restore(i_state);
}

/* ps() will show strange values for the stack usage of the main process.
   This is beacuse we do not fill the stack space of the main process
   with STACKPAT. The main processes has its stack in the real stack segment
   instead of in data space. It seems like a risky business to poke around
   outside of the data segment, and especially when you are currently using
   the stack whose contents you want to fill with STACKPAT.
   
   The IBM PC version checks the NULL address. Apparently this is is the
   bottom of the stack on the PC. This method is not useful on a UNIX
   machine however.
*/


/* Print process table info
 * Since things can change while ps is running, the ready proceses are
 * displayed last. This is because an interrupt can make a process ready,
 * but a ready process won't spontaneously become unready. Therefore a
 * process that changes during ps may show up twice, but this is better
 * than not having it showing up at all.
 */
int
ps(argc,argv,p)
int argc;
char **argv;
void *p;
{
	register struct proc *pp;
	int i;
	tprintf("PID       SP        stksize   maxstk    event     fl  in out name\n");
	for(i=0;i<PHASH;i++){
		for(pp = Waittab[i];pp != NULLPROC;pp = pp->next){
			if(tprintf("%-10lx%-10lx%-10u%-10u%-10lx%c%c%c %2d %2d  %s\n",
			 ptol(pp),0,pp->stksize,0,
			 ptol(pp->event),
			 pp->i_state ? 'I' : ' ',
			 (pp->state & WAITING) ? 'W' : ' ',
			 (pp->state & SUSPEND) ? 'S' : ' ',
			 pp->input,pp->output,
			 pp->name) == EOF)
				return 0;
		}
	}
	for(pp = Rdytab;pp != NULLPROC;pp = pp->next){
		if(tprintf("%-10lx%-10lx%-10u%-10u          %c%c%c %2d %2d  %s\n",
		 ptol(pp),0,pp->stksize,0,
		 pp->i_state ? 'I' : ' ',
		 (pp->state & WAITING) ? 'W' : ' ',
		 (pp->state & SUSPEND) ? 'S' : ' ',
		 pp->input,pp->output,
		 pp->name) == EOF)
			return 0;
	}
	if(Curproc != NULLPROC){
		tprintf("%-10lx%-10lx%-10u%-10u          %c   %2d %2d  %s\n",
		 ptol(Curproc),0,Curproc->stksize,0,
		 istate() ? 'I' : ' ',
		 Curproc->input,Curproc->output,
		 Curproc->name);
	}
	return 0;
}

void kinit()
{
}

phash(event)
void *event;
{
	register unsigned x;

	x = (unsigned) event;

	/* If PHASH is a power of two, this will simply mask off the
	 * higher order bits
	 */
	return x % PHASH;
}
